
#include <iostream>
#include <fstream>
#include <limits>
#include <mpi.h>

#include "superdirectivity.h"
#include "sph_bessel.h"
#include "jade.h"

using namespace std;

int main(int argc, char* argv[]) {
	int ni = 1; // number of spherical interfaces
	Real kr_max = Real(1); // maximum kr
	Real eps_max = Real(225); // maximum epsilon
	sphere_ml_parallel cls_sph; // class for calculation of directivity
	vector<Real> vb_min, vb_max; // parameter boundaries

		// boundaries for all optimization parameters:
	vb_min.push_back(Real(0.01)); vb_max.push_back(Real(1.5)*kr_max); // dipole position
	for (int i=0; i<ni; ++i)
		vb_min.push_back(Real(0.01)); vb_max.push_back(Real(1.5)*kr_max); // radii
	for (int i=0; i<ni; ++i)
		vb_min.push_back(Real(1)); vb_max.push_back(eps_max); // real permittivities

	ofstream out_opt("e:/Alexey/03 - code/data/directivity/opt.dat");

	MPI_Init(&argc, &argv);

	int rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

		// Settings for optimization algorithm;
	int dimension = 2*ni+1; /// Number of parameters to optimize.
	int total_population = 3*dimension;  /// Total number of individuals in population.
	jade::SubPopulation sub_population;
	sub_population.Init(total_population, dimension);
	sub_population.FitnessFunction_p = &get_directivity_sphml_dZx_epsRe_parallel;
		/// Low and upper bound for all dimensions;
		//sub_population.SetAllBoundsVectors({0.005, 0.005}, {0.05, 0.02});
	sub_population.SetAllBoundsVectors(vb_min, vb_max);
	//sub_population.SetTargetToMinimum();
	sub_population.SetTargetToMaximum();
	sub_population.SetTotalGenerationsMax(10000);
	sub_population.SetBestShareP(0.05);
	sub_population.SetAdapitonFrequencyC(0.1);
	sub_population.SetDistributionLevel(0);
	//sub_population.PrintParameters("f1");
	//sub_population.RunOptimization(out_opt, reinterpret_cast<void*>(&cls_sph));
	sub_population.RunOptimization(cout, reinterpret_cast<void*>(&cls_sph));
	sub_population.PrintResult("directivity function ");
	MPI_Finalize();

	return 0;
}